# -*- coding: utf-8 -*-
"""
Created on Mon Feb  2 14:50:41 2015

@author: crovetta
"""

import copy
import itertools as it
import numpy as np
import state

import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.gridspec as gridspec



# -------------------- Arc ----------------------
class Arc(object):

    """ Arc is the class of arcs diagram.
        An arc has an origin point and a destination point.

        Args:
        self.ori: origin point represented by a tuple
        self.des: destination point represented by a tuple
    """
    
    def __init__(self,o,d):

        if isinstance(o,int):
            self.ori=tuple([o])
            self.des=tuple([d])

        elif isinstance(o,list):
            self.ori=tuple(o)
            self.des=tuple(d)

        else:
            self.ori=o
            self.des=d

    def __eq__(self,a2):
        t1=(self.ori==a2.ori)
        t2=(self.des==a2.des)
        return(t1&t2)

    def __ne__(self,a2):
        t1=(self.ori!=a2.ori)
        t2=(self.des!=a2.des)
        return(t1|t2)
        
    def __le__(self,a2):
        
        #print self, ' <= ', a2 ,' ?'

        for z in xrange(0,len(self.ori)):
            if self.ori[z] > a2.ori[z]:
                return False
            if self.des[z] > a2.des[z]:
                return False
                
        return True
        
    def __hash__(self):
        big=self.ori+self.des
        return hash(big)

    def __str__(self):

        if len(self.ori)==1:
            s='Arc[{self.ori[0]},{self.des[0]}]'.format(self=self)
            return s
        
        s='Arc[{self.ori},{self.des}]'.format(self=self)
        return(s)

    def __repr__(self):

        if len(self.ori)==1:
            s='Arc[{self.ori[0]},{self.des[0]}]'.format(self=self)
            return s
        
        s='Arc[{self.ori},{self.des}]'.format(self=self)
        return(s)


    def value(self,z=0):
        
        """ Gets the arc value in z position """
        
        v= self.des[z] - self.ori[z]
        return v
        
        
    def growing(self):
        
        """ Saids if the arc is conform is growing """
        
        a=np.array(self.ori)
        b=np.array(self.des)
        c=b-a
        
        d=c[c<0]
        
        # Growing test
        if sum(d)<0:
            return False
        
        return True
            
        
        
    def conform(self,popk):
        
        """ Saids if the arc respects the capacity queue """
        
        a=np.array(self.ori)
        b=np.array(self.des)
        v=b-a
        
        # Visited test
        e=v>0
        f=np.logical_or(e,popk)
        test=(f==popk).all()
        
        return test
        
        
        

    def arcvalue(self):
        
        """ Returns the population vector represented by the arc """
        
        zmax=len(self.ori)

        if zmax==1:
            return self.value()
        
        a=np.array(self.ori)
        b=np.array(self.des)
        
        c=b-a

        return tuple(c)
        
    def is_empty(self):
        
        """ Saids if the arcs represents an empty queue"""
        
        a=np.array(self.ori)
        b=np.array(self.des)
        c=b-a
        
        sv=sum(c)
        
        return sv==0
        
    def is_full(self,C):
        
        """ Saids if the arcs represents a full queue"""
        
        v=self.arcvalue()
        sv=sum(v)
        
        return sv==C

    def move(self,do,dd,z=0):

        """ Returns a new arc: Arc[self.ori+do,self.des.dd]"""
        
        o=list(self.ori)
        d=list(self.des)
        o[z]+=do
        d[z]+=dd

        a=Arc(tuple(o),tuple(d))
        
        return a
        
    
    def get_z_LARGEST(self):
        
        """ Returns the largest class """
        
        a=np.array(self.ori)
        b=np.array(self.des)
        c=b-a
        
        v=tuple(c)
        
        z=v.index(max(v))
        
        return z
        
    def get_z_PRIORITY(self):
        
        """ Returns the service class """
        
        a=np.array(self.ori)
        b=np.array(self.des)
        v=b-a
        
        for z in xrange(0,len(self.ori)):
            if v[z]>0:
                return z
        
        return -1
        
    
    def get_z(self,dis='PRIORITY'):
        
        """ Returns the service class according to the discipline """
        
        if dis=='PRIORITY':
            return self.get_z_PRIORITY()
            
        if dis=='LARGEST':
            return self.get_z_LARGEST()
        
        return -1
        
        
# ------------------ Column ---------------------------------  

def makeColomn_1(C,M,k):

    """ Constructs the k-th colomn of a complete diagram - 1 class """

    omin=max(0,M-sum(C[k:]))
    omax=min(M,sum(C[:k]))
    dmin=max(0,M-sum(C[k+1:]))
    dmax=min(M,sum(C[:k+1]))

    if(k==0):
        A0={Arc(0,d) for d in xrange(dmin,dmax+1)}
        C0=Column(A0)
        return C0

    if(k==len(C)):
        AK={Arc(o,M) for o in xrange(omin,omax+1)}
        CK=Column(AK)
        return CK

    Ak={Arc(o,d) for d in xrange(dmin,dmax+1) for o in xrange(omin,omax+1) if o <= d and (d-o)<=C[k]}
    Ck=Column(Ak)
        
    return Ck
    

def makeDiagram_1(C,M):

    """ Constructs the complete diagram (one class) """

    K=len(C)
    D=[makeColomn_1(C,M,k) for k in xrange(K)]
    return D


def test_node(M,n):
    
    aM=np.array(M)
    an=np.array(n)
    
    c=aM>=an
    
    return sum(c)==len(M)
    
    

def get_itAllNode(M):
    
    """ Gets the iterator of all possible nodes in a column """
    
    itNmax=it.product(range(0,max(M)+1),repeat=len(M))
    itNode=(n for n in itNmax if test_node(M,n))
    
    
    return itNode

    

def makeDiagram_Z(Mod):
    
    """ Constructs the complete diagram according to the model """
    
    M=Mod.M
    K=Mod.K
    Z=Mod.Z
    KZ=Mod.KZ
    
    
    # ============= Nodes construct ===============
    # Mother node (0,...,0)
    mother=tuple(it.repeat(0,Z))
   
    # End node
    dau=tuple(M)
    
    # ============ Columns construct ============
    D=[]
    
    # First column
    L0=[]
    itNode0=get_itAllNode(M) # Set of destination node
    KZ0=KZ[0]
    for des in itNode0:
        a=Arc(mother,des)
        if a.conform(KZ0):
            L0.append(a)
    C0=Column(L0)
    D.append(C0)
    
   
    # Intermediate columns
    Lk=[]
    itNode1=get_itAllNode(M)
    for ori in itNode1:
        itNode2=get_itAllNode(M)
        for des in itNode2:
            a=Arc(ori,des)
            if a.growing():
                Lk.append(a)
    
    
    Ck=C0
    for k in xrange(1,K-1):
        KZk=KZ[k]
        Lkk=[ a for a in Lk if a.conform(KZk)]
        Ck=Column(Lkk)
        D.append(Ck)
        
        
        
    # Last column
    LK=[]
    itNodeK=get_itAllNode(M) # Set of destination node
    KZK=KZ[-1]
    for ori in itNodeK:
        a=Arc(ori,dau)
        if a.conform(KZK):
            LK.append(a)
    CK=Column(LK)
    D.append(CK)
        

    return D
      

class Column(set):
    
        def __str__(self):
            
            """ to print a Column"""
            
            la=list(self)
            la.sort()
            laf=[str(a) for a in la]
            aff='\n'.join(laf)
            aff='Column{\n'+aff+'\n}\n|Column|='+str(len(self))+'\n'
            return aff
            
            
        def get_From(self,Ori):
            
            """ Returns a column of arcs in self where arcs origin are in Ori"""
            
            Ao={a for a in self if a.ori in Ori}
            return Column(Ao)
            
            
            
        def get_To(self,Des):
            
            """ Returns a column of arcs in self where arcs destination are in Des"""
            
            Ad={a for a in self if a.des in Des}
            return Column(Ad)
            
        
        def get_Origins(self):

            """ Returns the set of all arcs origins """            
            
            O={a.ori for a in self}
            return O
            
            
        def get_Destinations(self):

            """ Returns the set of all arcs destinations """            
            
            D={a.des for a in self}
            return D
            
            
        def get_Empty(self):
            
            """ Returns a Column with the arcs with values 0 """
            
            E={ a for a in self if a.is_empty() }
            return Column(E)
            
            
        def move(self,do,dd,z=0):
            
            """ Moves all arcs in the columns """
            
            M={a.move(do,dd,z) for a in self}
            return Column(M)
            
            
            

class Diagram(object):
    
    def __init__(self,Mod):
        self.Model=Mod
        self.mother=tuple(it.repeat(0,self.Model.Z))
        
        # Makes the state space
        if self.Model.Z==1: #if one class
            self.diagram=makeDiagram_1(self.Model.C,self.Model.M)
        else:
            self.diagram=makeDiagram_Z(self.Model)
            self.diagram=self.get_subDiagram_1(0,self.Model.K-1,self.diagram[0])
            self.diagram=self.get_subDiagram_2(self.Model.K-1,0,self.diagram[-1])
            
        self.complete=copy.deepcopy(self.diagram)
        
        
    def __str__(self):
        sd=[str(c) for c in self.diagram]
        affD='\n'.join(sd)
        affD='\nD= [\n'+affD+'\n]\n|D|='+str(len(self))
        return affD
        
    def __len__(self):
        t=0
        for c in self.diagram:
            t+=len(c)
        return t
        
    def reset(self):
        
        """ Transphorms the diagram into a complete diagram """
        
        self.diagram=copy.deepcopy(self.complete)

    def onePath(self):

        """ Saids if the diagram contains only one path """
        
        for col in self.diagram:
            if len(col)>1:
                return False
        return True
        
        
    def plot_subDiagram(self,k,B,leg='$B$',co='#0040ff'):
        
        """ Plots Link(B,A) where B is a subset of Ak """
        
        # use latex for font rendering
        mpl.rcParams['text.usetex'] = True
        mpl.rcParams['font.size'] = 20
        
        DB=self.Link(k,B)
        
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        
        K=self.Model.K
        M1=self.Model.M[0]+1
        M2=self.Model.M[1]+1
        
        ax.set_xticks(range(0,K+1))
        ax.set_yticks(range(0,M1))
        ax.set_zticks(range(0,M2))
        
        lab=0
        for kk in xrange(0,K):
            x=[kk,kk+1]
            DBk=DB[kk]
            
            #print 'kk=',kk,'DBk=',DBk
            
            # Plot nodes
            if kk<K-1:
                Des=DBk.get_Destinations()
                for d in Des:
                    ax.scatter(x[1],d[0],d[1],s=100, c='w', marker='o')
                
            # Plot arcs
            for a in DBk:
                y=[a.ori[0], a.des[0]]
                z=[a.ori[1], a.des[1]]
                ax.plot(x,y,z,c='black')
                
                if kk==k:
                    if lab==0:
                        ax.plot(x,y,z,c=co,linewidth=5,label=leg)
                        lab=lab+1
                    else:
                        ax.plot(x,y,z,c=co,linewidth=5)
                    
        
        ax.legend(loc='best')
        #plt.title('$Link($'+leg+'$,A)$')
        ax.set_xlabel(r'$k$',fontsize=25)
        ax.set_ylabel(r'$z=1$',fontsize=25)
        ax.set_zlabel(r'$z=2$',fontsize=25)
            
        ax.scatter(0,0,0,s=100, c='red', marker='o') # Mother node
        ax.scatter(K,M1-1,M2-1,s=100, c='red', marker='o') # End node
            
        plt.show()
        
        
    def plot_Transition(self,k,dis='PRIORITY'):
        
        # use latex for font rendering
        mpl.rcParams['text.usetex'] = True
        mpl.rcParams['font.size'] = 20
        col=['#0000FF','#01DF01','#DF0101','#A901DB']
    
        """ Plots the column k (Z<2)"""
    
        if self.Model.Z ==1:
            
            M=self.Model.M
            C=self.diagram[k]
            
            x=[k,k+1]
            
            # Plot column k
            E=C.get_Empty()
            T=C-E
            
            plt.axis([k-1,k+2,-1,M+1])
            plt.title('Column '+str(k+1))
            plt.xlabel('K')
            plt.ylabel('M')
            
            
            
            # Nodes
            N=C.get_Origins()
            for n in N:
                plt.scatter(x[0],n,s=100, c='white', marker='o')
                
            N=C.get_Destinations()
            for n in N:
                plt.scatter(x[1],n,s=100, c='white', marker='o')
            
            
            cc=0
            # Empty
            for e in E:
                y=[e.ori, e.des]
                
                if cc==0:
                    plt.plot(x,y,c='blue',label='E')
                    cc=cc+1
                    
                else:
                    plt.plot(x,y,c='blue')
                
            #Transit
            for t in T:
                y=[t.ori, t.des]
                if cc==1:
                    plt.plot(x,y,'g--',label='NE')
                    cc=cc+1
                else:
                    plt.plot(x,y,'g--')
            
            plt.legend(loc='lower right')
            plt.show()
            
            
        if self.Model.Z ==2:
            
            M1=self.Model.M[0]+1
            M2=self.Model.M[1]+1
            
            C=self.diagram[k]
            E=C.get_Empty()
            T=C-E
            
            x=[k,k+1]
            

            # Nodes
            No=C.get_Origins()
            Nd=C.get_Destinations()
            
            mpl.rcParams['text.usetex'] = True
            
            # =========== ax1: Column ======================
            ax1 = plt.subplot2grid((1,4), (0, 0), projection='3d')
            plt.title('$A_'+str(k+1)+'$')
            
            plt.xticks(range(k,k+2))
            plt.yticks(range(0,M1))
            ax1.set_zticks(range(0,M2))
          
            
            for n in No:
                ax1.scatter(x[0],n[0],n[1],s=100, c='white', marker='o')         
            for n in Nd:
                ax1.scatter(x[1],n[0],n[1],s=100, c='white', marker='o')
            
            # Empty
            for e in E:
                y=[e.ori[0], e.des[0]]
                z=[e.ori[1], e.des[1]]
                ax1.plot(x,y,z,c=col[0])
                   
                
            #Transit
            CA=[[],[]]
            for t in T:
                y=[t.ori[0], t.des[0]]
                z=[t.ori[1], t.des[1]]
                ca=t.get_z(dis)
                CA[ca].append(t)
            
                ax1.plot(x,y,z,c=col[1])
                    
               
            ax1.set_xlabel('$k$',fontsize=25)
            ax1.set_ylabel('$z=1$',fontsize=25)
            ax1.set_zlabel('$z=2$',fontsize=25)
            
            #ax1.legend(loc='best')
            
        
           
           # =========== ax2: T[0] ======================
            ax0 = plt.subplot2grid((1,4), (0, 1),projection='3d')
            plt.title('$P[0]$')
            plt.xticks(range(k,k+2))
            plt.yticks(range(0,M1))
            ax0.set_zticks(range(0,M2))
             
            for n in No:
                ax0.scatter(x[0],n[0],n[1],s=100, c='white', marker='o')         
            for n in Nd:
                ax0.scatter(x[1],n[0],n[1],s=100, c='white', marker='o')
                
            for a in E:
                y=[a.ori[0], a.des[0]]
                z=[a.ori[1], a.des[1]]
                ax0.plot(x,y,z,c=col[0])
            
            ax0.set_xlabel('$k$',fontsize=25)
            ax0.set_ylabel('$z=1$',fontsize=25)
            ax0.set_zlabel('$z=2$',fontsize=25)
            
            # =========== ax2: T[1] ======================
            ax2 = plt.subplot2grid((1,4), (0, 2),projection='3d')
            plt.title('$P[1]$')
            plt.xticks(range(k,k+2))
            plt.yticks(range(0,M1))
            ax2.set_zticks(range(0,M2))
             
            for n in No:
                ax2.scatter(x[0],n[0],n[1],s=100, c='white', marker='o')         
            for n in Nd:
                ax2.scatter(x[1],n[0],n[1],s=100, c='white', marker='o')
                
            for a in CA[0]:
                y=[a.ori[0], a.des[0]]
                z=[a.ori[1], a.des[1]]
                ax2.plot(x,y,z,c=col[1])
            
            ax2.set_xlabel('$k$',fontsize=25)
            ax2.set_ylabel('$z=1$',fontsize=25)
            ax2.set_zlabel('$z=2$',fontsize=25)
            
            
                
            # =========== ax3: T[2] ======================
            ax3 = plt.subplot2grid((1,4), (0, 3),projection='3d')
            plt.title('$P[2]$')
            plt.xticks(range(k,k+2))
            plt.yticks(range(0,M1))
            ax3.set_zticks(range(0,M2))
             
            for n in No:
                ax3.scatter(x[0],n[0],n[1],s=100, c='white', marker='o')         
            for n in Nd:
                ax3.scatter(x[1],n[0],n[1],s=100, c='white', marker='o')
                
            for a in CA[1]:
                y=[a.ori[0], a.des[0]]
                z=[a.ori[1], a.des[1]]
                ax3.plot(x,y,z,col[1])
            
            ax3.set_xlabel('$k$',fontsize=25)
            ax3.set_ylabel('$z=1$',fontsize=25)
            ax3.set_zlabel('$z=2$',fontsize=25)
            
            plt.show()
            
            B=Column(set(CA[0]))
            self.plot_subDiagram(1,B,'$P[1]$',col[1])
            B=Column(set(CA[1]))
            self.plot_subDiagram(1,B,'$P[2]$',col[1])
            
           
            
    
    
        
    
    def plot(self):
        
        """ Plot diagram (if Z<=2) """
        
        if self.Model.Z==1:
            
            M=self.Model.M
            K=self.Model.K
            
            plt.axis([-1,K+1,-1,M+1])
            plt.grid()
            
            plt.scatter(0,0,s=80, c='red', marker='o') # Mother node
            
            for k in xrange(0,K):
                x=[k,k+1]
                C=self.diagram[k]
                
                # Plot nodes
                Des=C.get_Destinations()
                for d in Des:
                    plt.scatter(x[1],d,s=80, c='w', marker='o')
                
                # Plot arcs
                for a in C:
                    y=[a.ori, a.des]
                    plt.plot(x,y,c='black')
            
            plt.show()
            
            return 1
                
            
        if self.Model.Z==2:
            
            
            K=self.Model.K
            M1=self.Model.M[0]+1
            M2=self.Model.M[1]+1
            
            fig = plt.figure()
            
            mpl.rcParams['text.usetex'] = True
            
            ax = fig.add_subplot(111, projection='3d')
            
            ax.set_xticks(range(0,K+1))
            ax.set_yticks(range(0,M1))
            ax.set_zticks(range(0,M2))
           
           
            
            for k in xrange(0,K):
                x=[k,k+1]
                C=self.diagram[k]
                
                # Plot nodes
                
                if k<K-1:
                    Des=C.get_Destinations()
                    for d in Des:
                        ax.scatter(x[1],d[0],d[1],s=100, c='w', marker='o')
                        #ax.text(x[1],d[0],d[1],'lol', fontsize=15)
                    
                
                # Plot arcs
                for a in C:
                    y=[a.ori[0], a.des[0]]
                    z=[a.ori[1], a.des[1]]
                    
                    ax.plot(x,y,z,c='black')
                   
                   
                    
                    
            
            ax.set_xlabel('$k$',fontsize=25)
            ax.set_ylabel('$z=1$',fontsize=25)
            ax.set_zlabel('$z=2$',fontsize=25)
            
            ax.scatter(0,0,0,s=100, c='red', marker='o') # Mother node
            ax.scatter(K,M1-1,M2-1,s=100, c='red', marker='o') # End node
            
            plt.show()
            
            return 1
            
        
        return 0
    
        
    
    def card_psi(self):

        """ Returns the number of path in the diagram """

        # Initalisation : find number of path in last colomn
        l=[(a.ori,1) for a in self.diagram[-1]]
        Path1=dict(l)
        
        # Other colomn
        itk=reversed(xrange(0,self.Model.K -1))
        for k in itk:

            Dk=self.diagram[k]
            DkO=Dk.get_Origins()
            Path2={}

            for o in DkO:
                
                Ao=Dk.get_From(set([o]))

                lo=0
                for a in Ao:
                    vdes=Path1[a.des] # Destinations values
                    lo+=vdes
               
                Path2[o]=lo

            Path1=Path2
            
        nb=Path1[self.mother]
        
        return nb
        
    
    def psi(self):

        """ psi function """

        # Initalisation : find path in last colomn
        l=[(a.ori,[[a.arcvalue()]] ) for a in self.diagram[-1]]
        Path1=dict(l)
        
        # Other colomn
        itk=reversed(xrange(0,self.Model.K -1))
        for k in itk:

            Dk=self.diagram[k]
            DkO=Dk.get_Origins()
            Path2={}

            for o in DkO:
                
                Ao=Dk.get_From(set([o]))

                lo=[]
                for a in Ao:
                    Vdes=Path1[a.des] # Destinations values
                    loa=[ [a.arcvalue()]+v for v in Vdes]
                    lo.extend(loa)
               
                Path2[o]=lo

            Path1=Path2
            
        Path=Path1[self.mother]
        TPath=[tuple(p) for p in Path]

        
        return TPath
        
    
    def to_setOfState(self):
        
        """ Returns a set of states """
        
        SD=self.psi()
        SS=state.convert_inSetOfState(SD)
        
        return SS
        
        
    
    def get_subDiagram_1(self,i,j,SDi):
        
        """ Gets the subdiagram (SDi,D[i+1],..,D[j]) generated by SDi -> """
        
        SD=[SDi] # First column
        SDk=SDi

        for k in xrange(i+1,j+1):
            Dk=self.diagram[k] # Colunm to treat
            Ori=SDk.get_Destinations() # Gets origins
            SDk=Dk.get_From(Ori) # Gets the Column with a.ori in Ori
            
            # TO DO : si SDk==Dk -> tout copier
            
            SD.append(SDk)
        return SD
        
    
    def get_subDiagram_2(self,i,j,SDi):
        
        """ Gets the subdiagram (D[j],..,D[i-1],SDi) generated by SDi <- """
        
        SD=[SDi] # First column
        SDk=SDi

        it=reversed(xrange(j,i))
        for k in it:
            Dk=self.diagram[k] # Colunm to treat
            Des=SDk.get_Origins() # Gets origins
            SDk=Dk.get_To(Des) # Gets the Column with a.des in Des
            
            # TO DO : si SDk==Dk -> tout copier
            
            SD.append(SDk)
        
        SD.reverse()
        
        return SD
        
    
    def Link(self,k,B):
        
        """ Gets Link(A,B) """
        
        K=self.Model.K
        
        D1=self.get_subDiagram_2(k,0,B) # Columns <=k
        D2=self.get_subDiagram_1(k,K-1,B) # Columns >k
        D3=D1+D2[1:]
        
        return D3
        
        
    
        
        
        
    
    def T1(self,i,j):
        
        """ Transition i->j then Z=1 """
        
        Di=self.diagram[i] # Column i
        Ei=Di.get_Empty() # Empty[i]
        Ti=Di-Ei # Transit[i]
            
        if i<j: # --->
        
            Empty=self.get_subDiagram_1(i,j,Ei)
            Transit=self.get_subDiagram_1(i,j,Ti)
                
                
            # k=i
            Ti=Transit[0]
            Ti=Ti.move(0,-1)
            self.diagram[i]=Ti.union(Empty[0])
                
            # i < k < j
            m=1
            for k in xrange(i+1,j):
                Tk=Transit[m]
                Tk=Tk.move(-1,-1)
                self.diagram[k]=Tk.union(Empty[m])
                m+=1
                
            # k = j
            Tj=Transit[-1]
            Tj=Tj.move(-1,0)
            self.diagram[j]=Tj.union(Empty[-1])
                
                
            
        if i>j: # <---
                
            Empty=self.get_subDiagram_2(i,j,Ei)
            Transit=self.get_subDiagram_2(i,j,Ti)
                
            # k=i
            Ti=Transit[-1]
            Ti=Ti.move(1,0)
            self.diagram[i]=Ti.union(Empty[-1])
           

            # j < k < i
            m=1
            for k in xrange(j+1,i):
                Tk=Transit[m]
                Tk=Tk.move(1,1)
                self.diagram[k]=Tk.union(Empty[m])
                m+=1
                
                
            # k=j
            Tj=Transit[0]
            Tj=Tj.move(0,1)
            self.diagram[j]=Tj.union(Empty[0])
            
                
        
    
            
    
    def T(self,i,J,dis='PRIORITY'):
        
        """ Transition on diagram """
        
        # =========== Case Z=1 =================
        if self.Model.Z==1:
            j=J 
            self.T1(i,j)
            
            return 1
            
        
        
        Di=self.diagram[i]
        
        # Empty
        Ei=Di.get_Empty()
        Empty=self.Link(i,Ei)
        
        # Split column i
        Ti=Di-Ei
        DicTz={}
        itTi=(a for a in Ti)
        for a in itTi:
            z=a.get_z(dis) # Arc a choose its service class
            
            if DicTz.has_key(z)==False:
                DicTz[z]=[]
            
            La=DicTz[z]
            La.append(a)            
            
        
        # Compute Transit 
        Transit={}
        for z in DicTz.keys():
            
            Tiz=DicTz[z]
            Ciz=Column(Tiz)
            Transit[z]=self.Link(i,Ciz)
            
            #txt='T['+str(z+1)+']'
            #self.plot_subDiagram(i,Ciz,txt)
            
        
        # Compute Transit'
        Move={}
        for z in DicTz.keys():
            
            D=Transit[z]
            j=J[z]
            
            #print 'Transition',i,'->',j
            
            if i<j:
                
                # Column i
                Col=D[i]
                D[i]=Col.move(0,-1,z)
                
                # Column i<k<j
                for k in xrange(i+1,j):
                    Col=D[k]
                    D[k]=Col.move(-1,-1,z)
                
                # Column j
                Col=D[j]
                D[j]=Col.move(-1,0,z)
                
                Move[z]=range(i,j+1)
                
                
            if j<i:
                
                # Column i
                Col=D[i]
                D[i]=Col.move(1,0,z)
                
                # Column j<k<i
                for k in xrange(j+1,i):
                    Col=D[k]
                    D[k]=Col.move(1,1,z)
                
                # Column j
                Col=D[j]
                D[j]=Col.move(0,1,z)
                
                Move[z]=range(j,i+1)
                
            
            
            
        # Compute Empty+Transit[0]+Transit[1]...
        for k in xrange(0,self.Model.K):
            C=Empty[k]
            for z in DicTz.keys():
               Cz=Transit[z][k]
               C=C.union(Cz)
                    
            self.diagram[k]=C
                
            

            
           
            
            
        
            
            
                
                
               
            
            
            
        
            
        
            
                
               
                
            
            
            
            
    




