# -*- coding: utf-8 -*-
"""
Created on Wed Jan 28 15:14:19 2015

@author: C. Rovetta
"""

import random
import numpy as np
import scipy.special as sp
from diagram import Diagram


def make_proba(L):
    
    """ Returns a probability tuple"""
    
    total=sum(L)
    P=[]
    
    P.append(float(L[0])/total)
    for i in xrange(1,len(L)):
        pi=P[i-1]+(float(L[i])/total)
        P.append(pi)
    
    return P
    
    
    
    
class ClassRouting(object):
    
    def __init__(self,Mat):
        self.Mat=Mat
        self.PM=np.cumsum(Mat,axis=1)
        
        
        vK=self.PM[:,-1].astype(bool)
        self.vK=vK # Queue present in the network
        
    def __repr__(self):
        aff='\n Routing matrix:\n '+str(self.Mat)+'\n Queues in class-network: '+str(self.vK)+'\n'
        return aff
        
    def queue_number(self):
        vi=self.vK.astype(int)
        qn=sum(vi)
        return qn
        
        
    def rand_j(self,i):
        
        """ Find a random destination j """
        
        if self.in_network(i)==False:
            return -1
        
        Pi=self.PM[i,:]
        K=len(Pi)
        
        nbj=random.random()
        k=0
        while (k < K)&(nbj >= Pi[k]):
            k=k+1
        
        return k
        
    def in_network(self,k):
        return(self.vK[k])
    
    
        


class Model(object):
    
    
    
    def __init__(self,M,Mu,R,C=0):
        
        # ===== Polulation ========
        self.M=M #Population vector
        
        # ====== Queues ===========
        self.Mu=Mu #Service
        self.PMu=make_proba(Mu)
        self.K=len(Mu) # Queue number
        
        # == Routing propability =====
        self.R=R
        
        
        
        # One class model
        if isinstance(M,int):
            self.Z=1
            self.sM=M
            self.KZ=np.transpose(self.R[0].vK)
            
            if C == 0:
                self.C=tuple([M for k in xrange(0,self.K)])
            else:
                self.C=C[0:self.K]
            
        
        # Z>1 class model
        else:
            self.Z=len(M) #Class number
            self.sM=sum(M) # Total number of customers |M|
            
            
            # == Find the max vector in each queue ==
            CC=[]
            for z in xrange(0,self.Z):
                Pz=self.R[z]
                Vz=Pz.vK
                Vz=Vz.astype(int)
                Vz=Vz*self.M[z]
                Vz=list(Vz)
                CC.append(Vz)
            Q=np.array(CC)
            Q=np.transpose(Q)
            QC=list(Q)
            QC=[tuple(q) for q in QC]
            self.C=QC #Capacity vector - infinite capacity
            
            # == Table
            t1=[self.R[0].vK]
            for z in xrange(1,self.Z):
                t2=[self.R[z].vK]
                t1=np.concatenate((t1,t2),axis=0)
            t1=np.transpose(t1)
            self.KZ=t1
          
            
            
    def set_popVector(self,M):
        
        """ Modifies the population vector """
        
        self.M=M
        
    def class_in_queue(self,z,k):
        
        """ Saids if queue k is visited by the class z """
        
        return self.R[z].in_network()
        
        
        
    def statespace_size(self):
        
        """ Gets the customer number if all queue has infinite capacity """
        
        if self.Z==1:
            sz=sp.binom(self.K-1+self.M,self.M)
            return int(sz)
        
        sS=1
        for z in xrange(0,self.Z):
            Pz=self.R[z]
            Kz=Pz.queue_number()
            sz=sp.binom(Kz-1+self.M[z],Kz-1)
            sS=sS*sz
        
        print '|S|=',sS
        
        return sS
        
    def diagram_exat_size(self):
        
        """ Gets the arc number """
        
        D=Diagram(self)
        return(len(D))
        
        
    
    def diagram_size(self):
        
        """ Gets the arc number """
        
        #D=Diagram(self)
        
        
        
        # Compute |A|
        M=self.M
        MStar=1
        for m in M:
            MStar=(m+1)*MStar
            
        Ms=[((m+1)*(m+2))/2 for m in M]
        p=1
        for m in Ms:
            p=p*m
        
       
        esti=2*MStar+p*(self.K-2)
        
        #print '|psi(D)|=', D.card_psi()
        print '|A|<=',esti
        #print '|A|=',len(D)
        
        return(esti)
        
        
    def __str__(self):
        
        aff="\n\n ================= Model ======================"
        aff+='\n  Number of queues: K='+str(self.K)
        aff+='\n  Service rate: Mu='+str(self.Mu)
        aff+='\n  Capacity vector: C='+str(self.C)
        aff+='\n  Number of classes: Z='+str(self.Z)
        aff+='\n  Population vector: M='+str(self.M)+' |M|='+str(self.sM)
        #aff+='\n PMu='+str(self.PMu)
        aff+='\n  Network: \nR='+str(self.R)
        #aff+='\n  KZ= \n'+str(self.KZ)
        aff+='\n  Size of state space: '+str(self.statespace_size())
        aff+="\n ================================================"
       
        return aff
        
        
    def rand_i(self):
        
        """ Gets a random queue according to the service rate """
        
        nbi=random.random()
        k=0
        while (k < self.K)&(nbi >= self.PMu[k]):
            k=k+1
        i=k
        
        return i
        
        
    def rand_j(self,i,z=0):
        
        Pz=self.R[z]
        j=Pz.rand_j(i)
        
        return j
        
        
    def rand_iJ(self):
        
        """ Produces a random couple (i,J) """
        
        i=self.rand_i()
        J=self.rand_J(i)
        
        return (i,J)
    
    
    
    def rand_J(self,i):
        
        """ For all class gets j"""
        
        J=[]
        for z in xrange(0,self.Z):
            Pz=self.R[z]
            jz=Pz.rand_j(i)
            J.append(jz)
            
        if self.Z==1:
            return J[0]
        
        return tuple(J)
        
        
    
    
        
    
    
        
        
        
        




